/* Checks if the descent direction is computed correctly using the
   analytic solution; also checks if the algorithm converges
   to the right solution regardless of where it starts */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <cassert>

#include <dune/common/shared_ptr.hh>

#include <dune/tectonic/ellipticenergy.hh>

#include "test-gradient-method-helper.hh"
#include "test-gradient-method-nicefunction.hh"

int main() {
  int const dim = 2;
  typedef Dune::EllipticEnergy<dim> Functional;
  typedef Functional::SmallMatrix SmallMatrix;
  typedef Functional::SmallVector SmallVector;

  SmallMatrix A;
  A[0][0] = 3;
  A[0][1] = A[1][0] = 1.5;
  A[1][1] = 4;
  SmallVector b;
  b[0] = 1;
  b[1] = 2;

  // |x|^2 as the nonlinearity is the same as having no nonlinearity
  // but twice the identity matrix added to A. In other words, we're
  // solving A + 2*id = b
  auto const f = Dune::make_shared<Dune::Parabola const>();
  auto const phi = Dune::make_shared<Functional::NonlinearityType const>(f);
  Functional J(A, b, phi);

  SmallVector solution; // Analytic solution
  solution[0] = 4.0 / 37.0;
  solution[1] = 34.0 / 111.0;

  SmallMatrix M = A;
  M[0][0] += 2;
  M[1][1] += 2;

  double ret1;
  {
    SmallVector start = b;
    start *= 17;

    SmallVector analytic_descent = b;
    M.mmv(start, analytic_descent);

    SmallVector numerical_descent;
    J.descentDirection(start, numerical_descent);
    assert(two_distance<dim>(numerical_descent, analytic_descent) < 1e-10);

    ret1 = functionTester(J, start, 6);
    assert(two_distance<dim>(start, solution) < 1e-6);
  }

  double ret2;
  {
    // Something random
    SmallVector start;
    start[0] = 279;
    start[1] = -96;

    SmallVector analytic_descent = b;
    M.mmv(start, analytic_descent);

    SmallVector numerical_descent;
    J.descentDirection(start, numerical_descent);
    assert(two_distance<dim>(numerical_descent, analytic_descent) < 1e-10);

    ret2 = functionTester(J, start, 15);
    assert(two_distance<dim>(start, solution) < 1e-6);
  }
  assert(std::abs(ret1 - ret2) < 1e-11);
}
